# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .base import BasePlot, BasePlotOptions
from .manager import PlotManager
from .plotwidget import (
    PlotDialog,
    PlotOptions,
    PlotWidget,
    PlotWindow,
    SubplotWidget,
    SyncPlotWindow,
    SyncPlotDialog,
    set_widget_title_icon,
)
