# -*- coding: utf-8 -*-

"""
Cross-sections
^^^^^^^^^^^^^^

The cross-section panels are used to display cross-sections of 2D data, either
along the X or Y axis, or along an oblique rectangle (average of the data along
an oblique rectangle).

.. autoclass:: XCrossSection
.. autoclass:: YCrossSection
.. autoclass:: ObliqueCrossSection
.. autoclass:: LineCrossSection
"""

# pylint: disable=unused-import
# flake8: noqa

from .cswidget import (
    ObliqueCrossSection,
    XCrossSection,
    YCrossSection,
    LineCrossSection,
)
