# -*- coding: utf-8 -*-

"""
Colormaps
=========

The `colormap` module contains definition of common colormaps and tools
to manipulate and create them.

This package provides widgets for creating, editing, and managing colormaps:

- ``ColorMapWidget``: Interactive colormap editor widget
- ``ColorMapEditor``: Full colormap editing dialog
- ``ColorMapManager``: Colormap management interface
"""
